/*
 *  $Id: comm_split.c,v 1.7 1993/11/12 12:31:42 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */


#include <mpi.h>


typedef struct _MPIR_Table_Entry {
  int color;
  int key;
  int next;
} MPIR_Table_Entry;
#define MPIR_TABLE_ENTRY_NUM_ELEMENTS 3

/*@

MPI_Comm_split - creates new communicators based on colors and keys

algorithm:

The current algorithm used has quite a few (read: a lot of) inefficiencies 
that can be removed.  Here's what we do for now:
. 1) A table is built of colors, and keys (has a next field also).
. 2) The tables of all processes are merged using a "reduceall"-type process
. 3) Two contexts are allocated for all the comms to be created.  These
     same two contexts can be used for all created communicators since
     the communicators will not overlap.
. 4) If the local process has a color of MPI_UNDEFINED, it can return
     a NULL comm. 
. 5) The table entries that match the local process color are sorted 
     by key/rank. 
. 6) A group is created from the sorted list and a communicator is created
     with this group and the previously allocated contexts.
@*/

int MPI_Comm_split ( comm, color, key, comm_out )
MPI_Comm comm;
int color, key;
MPI_Comm *comm_out;
{
  int               size, rank, head, new_size, i;
  MPIR_Table_Entry *table, *table_buf;
  int               context;
  int              *group_list;
  MPI_Group         comm_group, group;
  MPI_Comm  new_comm;

  /* If we don't have a communicator we don't have anything to do */
  if ( (comm            == (MPI_Comm)0) ||
       (comm->comm_type == MPIR_INTER) ) {
    (*comm_out) = (MPI_Comm)0;
    return (MPI_SUCCESS);
  }

  /* Create and initialize split table.  We allocate space for two */
  /* tables since we need space to receive tables. */
  (void) MPI_Comm_size ( comm, &size );
  (void) MPI_Comm_rank ( comm, &rank );
  table = (MPIR_Table_Entry *) CALLOC ( 2 * size, sizeof(MPIR_Table_Entry) );
  table_buf         = table + size;
  table[rank].color = color;
  table[rank].key   = key;  
  table[rank].next  = 0;

  /* The two synchronizations in this function come when we combine the */
  /* tables and get the two contexts.  We could do these two things */
  /* in one step .... eventually */

  /* Combine the split table.  When we write the reduceall function */
  /* we can do away with this function */
  (void) MPIR_Combine_split_table ( comm, size, rank, table, table_buf );

  /* Allocate 2 contexts */
  (void) MPIR_Context_alloc( comm, 2, &context );

  /* If my color is MPI_UNDEFINED, then I'm not in a comm and can */
  /* stop here since there are no more communications with others */
  /* I'll even go ahead and free the 2 contexts I allocated above */
  if ( table[rank].color == MPI_UNDEFINED ) {
    FREE(table);
    (void) MPIR_Context_dealloc( comm, 2, context );
    (*comm_out) = (MPI_Comm)0;
    return (MPI_SUCCESS);
  }

  /* Sort the table */
  (void) MPIR_Sort_split_table ( size, rank, table, &head, &new_size );
     
  /* Create group of processes that share my color */
  group_list = (int *) MALLOC ( new_size * sizeof(int) );
  for ( i=0; i<new_size; i++, head=table[head].next )
    group_list[i] = head;
  (void) MPI_Comm_group ( comm, &comm_group );
  (void) MPI_Group_incl ( comm_group, new_size, group_list, &group );
  FREE(group_list);

  /* Make communicator using contexts allocated */
  new_comm                = (*comm_out) = NEW(MPIR_COMMUNICATOR);
  new_comm->comm_type     = MPIR_INTRA;
  (void) MPIR_GROUP_dup( group, &(new_comm->group) );
  new_comm->send_context  = new_comm->recv_context = context;
  new_comm->comm_cache    = 0;
  new_comm->topology.type = MPI_UNDEFINED;
  (void) MPIR_HBT_new_tree ( &(new_comm->attr_cache) );
  (void) MPIR_Comm_make_coll( new_comm );

  return (MPI_SUCCESS);
}


/*+

MPIR_Sort_split_table - sort split table using a yuckie sort (YUCK!).  I'll
                        switch to a more efficient sort one of these days.

+*/
#define MPIR_EOTABLE -1
int MPIR_Sort_split_table ( size, rank, table, head, list_size )
int size, rank;
MPIR_Table_Entry *table;
int *head, *list_size;
{
  int i, j, prev;
  int color = table[rank].color;
  
  /* Initialize head and list size */
  (*head)      = MPIR_EOTABLE;
  (*list_size) = 0;

  /* Sort out only colors == to my rank's color */
  for ( i=0; i<size; i++ ) {
    for (prev = MPIR_EOTABLE,j=(*head);j!=MPIR_EOTABLE;prev=j,j=table[j].next) {
      if ( table[i].color != color )
        continue;
      if ( (j==MPIR_EOTABLE) || (table[i].key<table[j].key) )
	break;
    }
    if ( table[i].color == color) {
      (*list_size)++;
      table[i].next = j;
      if (prev == MPIR_EOTABLE)
        (*head) = i;
      else
        table[prev].next = i;
    }
  }
}

/*+

MPIR_Add_split_tables - temporary function that will dissapear when
                        MPI_allreduce is available

+*/
int MPIR_Add_split_tables ( size, table_inout, table_in )
int size;
MPIR_Table_Entry *table_inout, *table_in;
{
  int i;
  for ( i=0; i<size; i++ ) {
    table_inout[i].color += table_in[i].color;
    table_inout[i].key   += table_in[i].key;
  }
  return (MPI_SUCCESS);
}

/*+

MPIR_Combine_split_table - temporary function that will dissapear when
                           MPI_reduceall is available

+*/
int MPIR_Combine_split_table ( communicator, size, rank, table, table_buf )
MPI_Comm communicator;
int size, rank;
MPIR_Table_Entry *table, *table_buf;
{
  int num_ints = size * MPIR_TABLE_ENTRY_NUM_ELEMENTS;
  int N2_prev, surfeit;
  int d, dst, src;

  /* If there's only one member, this is trivial */
  if ( size > 1 ) {

    /* Initialize variables */
    MPI_Comm_request s_req, r_req;
    MPI_Status       status;
    MPI_Comm comm = communicator->comm_coll;

    (void)MPIR_Comm_N2_prev ( comm, &N2_prev );
    surfeit = size - N2_prev;

    /* Combine the split tables from all processes */
    if ( rank < N2_prev ) {
      if( rank < surfeit ) {

	/* get the fanin letter from the upper "half" process: */
	dst = N2_prev + rank;
	
        MPI_Irecv( table_buf, num_ints, MPI_INT, dst, 0, comm, &r_req );
        MPI_Wait( r_req, &status );
	
	/* combine table */
        (void) MPIR_Add_split_tables ( size, table, table_buf );
      }

      /* combine on embedded N2_prev power-of-two processes */
      for (d = 1; d < N2_prev; d <<= 1) {
        dst = (rank ^ d);

        /* Send and receive a context limit */
        if ( rank < dst ) {
          MPI_Isend( table, num_ints, MPI_INT, dst, 0, comm, &s_req );
          MPI_Wait( s_req, &status );
          MPI_Irecv( table_buf, num_ints, MPI_INT, dst, 0, comm, &r_req );
          MPI_Wait( r_req, &status );
        }
        else {
          MPI_Irecv( table_buf, num_ints, MPI_INT, dst, 0, comm, &r_req );
          MPI_Wait( r_req, &status );
          MPI_Isend( table, num_ints, MPI_INT, dst, 0, comm, &s_req );
          MPI_Wait( s_req, &status );
        }
	
	/* combine table */
        (void) MPIR_Add_split_tables ( size, table, table_buf );
      }

      /* fanout data to nodes above N2_prev... */
      if ( rank < surfeit ) {
	dst = N2_prev + rank;
        MPI_Isend( table, num_ints, MPI_INT, dst, 0, comm, &s_req );
        MPI_Wait( s_req, &status );
      }
    } 
    else {
      /* fanin data to power of 2 subset */
      src = rank - N2_prev;
      MPI_Isend( table, num_ints, MPI_INT, src, 0, comm, &s_req );
      MPI_Wait( s_req, &status );
      MPI_Irecv( table, num_ints, MPI_INT, src, 0, comm, &r_req );
      MPI_Wait( r_req, &status );
    }
  } 
  return(MPI_SUCCESS); 
}






